import argparse
from Bio import SeqIO, bgzf
from Bio.Seq import Seq
from .util import _open, _open_out
import os

def fastq(input = None, output = None, unique = None, join = None, filter_q = None, rc = None, nproc = 1):
    """fastq subcommand
    Paramters
    ---------

    input : str
        input file name, auto detects .gz
    output : str
        output file name, auto detects .gz
    unique : str
        output unique fastq records by record ID (some bam coverted fastq may have records with duplicated IDs)
    join : str
        fastq file name to join to the input fastq, auto detects .gz
            "join" means to append each read to corresponding lines in input, not "cat"
    filter_q : int
        split reads into reads with mean_quality higher than filter_q, also output the reads failed to pass the filter into no_pass_xxx.fastq.gz.
    nproc: int
        number of parallel jobs, default to 1
        funtions that support nproc: rc

    # TODO:
    rc : str
        boolean, to reverse complement the reads, supports multi-processing

    """

    args = list(locals().keys())

    local = locals()
    if all(bool(local[key]) is not True for key in args): # use True since args can be either None or False
        print("scutls fastq: warning: use 'scutls fastq -h' for usage")

    # make fastq file unique if "-u":
    if unique:
        print("Saving to " + output + " ...")

        if not os.path.dirname(output) == "":
            os.makedirs(os.path.dirname(output), exist_ok=True)
 
        with _open(input) as f:
            unique_records = {}
            for record in SeqIO.parse(f, 'fastq'):
                if record.name not in unique_records:
                    unique_records[record.name] = record

            if output.endswith(".gz"):
                with bgzf.BgzfWriter(output, "wb") as outgz:
                    SeqIO.write(sequences = unique_records.values(), handle = outgz, format="fastq")
            else:
                SeqIO.write(unique_records.values(), output, "fastq")

        print("Done!")

    # join fastq file to input fastq if "-j":
    if join:
        print("Saving to " + output + " ...")

        if not os.path.dirname(output) == "":
            os.makedirs(os.path.dirname(output), exist_ok=True)

        f_input = _open(input)
        f_join  = _open(join)

        # sanity check: read count must match
        record_i = [record for record in SeqIO.parse(f_input, "fastq")]
        record_j = [record for record in SeqIO.parse(f_join, "fastq")]
        assert len(record_i) == len(record_j), "Number of records in --input fastq does not equal to that in --join fastq!"

        if output.endswith(".gz"):
            f_out = bgzf.BgzfWriter(output, "wb")
        else:
            f_out = open(output, "w")

        for i in range(len(record_i)):
            _record = record_i[i]
            _record_j = record_j[i]

            # sanity check: read description must match
            assert _record.description == _record_j.description, "Read description in --input fastq does not match with --join fastq!"
            record_quality = ''.join(map(lambda x: chr(x + 33), _record.letter_annotations["phred_quality"]))
            record_j_quality = ''.join(map(lambda x: chr(x + 33), _record_j.letter_annotations["phred_quality"]))

            f_out.write("@" + _record.description + "\n")
            f_out.write(str(_record.seq) + str(_record_j.seq) + "\n")
            f_out.write("+\n")
            f_out.write(record_quality + record_j_quality + "\n")

        f_input.close()
        f_join.close()
        f_out.close()
        print("Done!")

    # filter reads based on mean quality cutoff if "-fq":
    if filter_q:
        filter_q = int(filter_q)
        if not os.path.dirname(output) == "":
            os.makedirs(os.path.dirname(output), exist_ok=True)
        
        output2 = os.path.join(os.path.dirname(output), "not_pass_" + os.path.basename(output))
        print("Saving to " + output + " ...")
        print("Saving to " + output2 + " ...")

        with _open(input) as f, _open_out(output) as f_out, _open_out(output2) as f_out2:
            for record in SeqIO.parse(f, "fastq"):
                mean_quality = sum(record.letter_annotations["phred_quality"]) / len(record)
                if mean_quality >= filter_q:
                    SeqIO.write(record, f_out, "fastq")
                else:
                    SeqIO.write(record, f_out2, "fastq")
                    
        print("Done!")

    
    # reverse complement fastq if "-rc":
    # todo

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--input',  '-i', type   = str)
    parser.add_argument('--output', '-o', type   = str)
    parser.add_argument('--unique', '-u', action = 'store_true')
    parser.add_argument('--join',   '-j', type   = str)

    args = parser.parse_args()
    fastq(args.input, args.output, args.unique, args.join)

if __name__ == "__main__":
    main()



import argparse
from multiprocessing import Pool
import functools
from .util import bam_chunk_interval, bam_locate_pos_in_read, bam_locate_pos_in_ref

def bam(input = None, output = None, locate_pos_in_read = None, locate_pos_in_ref = False, nproc = 1):
    """bam subcommand
    Paramters
    ---------

    input : str
        input file name, must end with .bam
    output : str
        output file name
    locate_pos_in_read : int
        a reference coornidate to be located in each read in the bam file
        ------------------
        Ref:    AATCCGGC
        Query:  AA  CGGC
        Q-pos:  12  3456
        ------------------
        For the above case, if trying to locate reference coordinate 2 (0-based), which is "T", the query pos 2 will be returned. 
    locate_pos_in_ref : bool
        for each read in bam: for every cigar tuple, output the corresponding reference coordinate
    
    nproc : int
        number of parallel jobs, default to 1
        funtions that support nproc: locate_pos_in_read

    """
    args = list(locals().keys())

    local = locals()
    if all(bool(local[key]) is not True for key in args): # use True since args can be either None or False
        print("scutls bam: warning: use 'scutls bam -h' for usage")

    # locate a given reference coordinate in each read
    if locate_pos_in_read:
        if output == None:
            pass
        else:
            print("Processing ...")
            
        intervals = bam_chunk_interval(bam = input, nproc = nproc)
        p = Pool(nproc)
        res = p.map_async(
            functools.partial(
                bam_locate_pos_in_read,
                bam = input,
                ref_coordinate = int(locate_pos_in_read)),
                intervals.values()).get()

        res_locate = []
        for x in range(len(res)):
            res_locate += res[x]
        
        print(res_locate)
        
        if output == None:
            for x in res_locate:
                print(x)
        else:
            with open(output, "w") as f:
                for x in res_locate:
                    f.write(x + "\n")
            print("Done!")

    # locate each cigar tuple in reference coordinate for every read
    if locate_pos_in_ref:
        if output == None:
            pass
        else:
            print("Processing ...")
            
        intervals = bam_chunk_interval(bam = input, nproc = nproc)
        p = Pool(nproc)
        res = p.map_async(
            functools.partial(
                bam_locate_pos_in_ref,
                bam = input),
                intervals.values()).get()

        res_locate = []
        for x in range(len(res)):
            res_locate += res[x]
        
        if output == None:
            for x in res_locate:
                print(x)
        else:
            with open(output, "w") as f:
                for x in res_locate:
                    f.write(x + "\n")
            print("Done!")

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--input',  '-i', type = str)
    parser.add_argument('--output', '-o', type = str, default = None)
    parser.add_argument('--locate_pos_in_read', '-lpir', default = 0)
    parser.add_argument('--num_processor', '-nproc', default = 1, type = int)

    args = parser.parse_args()
    bam(args.input, args.output, args.locate_pos_in_read, args.num_processor)

if __name__ == "__main__":
    main()



import urllib3
from bs4 import BeautifulSoup
import re
import importlib_resources
import json
import gzip
from mimetypes import guess_type
from functools import partial
import math
from Bio import SeqIO
from Bio.Seq import Seq
import regex
import os 
import pysam

# helper function to perform sort
def num_sort(test_string):
    return list(map(int, re.findall(r'\d+', test_string)))[0]

# helper for extracting links to releases:
def get_links(url, http):
    r = http.request('GET', url)
    soup = BeautifulSoup(r.data, 'html.parser')
    a = soup.find_all('a', href = re.compile(r'^[^.]'))
    releases = [sp_h['href'].rstrip("/") for sp_h in a]
    releases = [release for release in releases if release.startswith("release-")]
    releases.sort(key=num_sort, reverse = True)
    return(releases)

# helper for extracting genome download url links:
def get_download_urls(http, ftp_site, release, genome_download_links):
    ftp_site_fasta = ftp_site + "release-" + release + "/fasta/"
    r = http.request('GET', ftp_site_fasta)
    soup = BeautifulSoup(r.data, 'html.parser')
    spcies_href = soup.find_all('a', href = re.compile(r'^[^.]'))

    species = [sp_h['href'].rstrip("/") for sp_h in spcies_href]

    for sp in species:
        genome_download_links.setdefault(sp, {})
        for k in ["gtf_md5sum", "genome_md5sum", "gtf", "genome"]:
            genome_download_links[sp].setdefault(k, None)

    species_link = [ftp_site_fasta + sp_h['href'] + "dna/" for sp_h in spcies_href]

    for i, sp_link in enumerate(species_link):
        req = http.request('GET', sp_link)
        soup = BeautifulSoup(req.data, 'html.parser')
        primary_seq_href = soup.find_all('a', href = re.compile(r'dna.primary_assembly.fa.gz'))
        toplevel_seq_href = soup.find_all('a', href = re.compile(r'dna.toplevel.fa.gz'))
        md5sum_href = soup.find_all('a', href = re.compile(r'CHECKSUMS'))

        if md5sum_href:
            genome_download_links[species[i]]["genome_md5sum"] = sp_link + md5sum_href[0]['href']
        if primary_seq_href:
            genome_download_links[species[i]]["genome"] = sp_link + primary_seq_href[0]['href']
        elif toplevel_seq_href:
            genome_download_links[species[i]]["genome"] = sp_link + toplevel_seq_href[0]['href']

    ## get gtf and gtf md5sum
    ftp_site_gtf = ftp_site + release + "/gtf/"
    r = http.request('GET', ftp_site_gtf)
    soup = BeautifulSoup(r.data, 'html.parser')
    spcies_href = soup.find_all('a', href = re.compile(r'^[^.]'))
    species_link = [ftp_site_gtf + sp_h['href']  for sp_h in spcies_href]

    species = [sp_h['href'].rstrip("/") for sp_h in spcies_href]
    for i, sp_link in enumerate(species_link):
        req = http.request('GET', sp_link)
        soup = BeautifulSoup(req.data, 'html.parser')
        gtf_href = soup.find_all('a', href = re.compile(r'\d+.gtf.gz'))
        md5sum_href = soup.find_all('a', href = re.compile(r'CHECKSUMS'))
        if md5sum_href:
            genome_download_links[species[i]]["gtf_md5sum"] = sp_link + md5sum_href[0]['href']
        if gtf_href:
            genome_download_links[species[i]]["gtf"] = sp_link + gtf_href[0]['href']
    return genome_download_links

def update_ensembl_release():
    """
    Query ENSEMBL website and update assets/genome_ensembl_release_all.txt.
    """

    http = urllib3.PoolManager()

    # get vetebrate releases: http://ftp.ensembl.org/pub/
    url = "http://ftp.ensembl.org/pub/"
    vet_releases = get_links(url, http)

    # get plant releases:
    url = "http://ftp.ensemblgenomes.org/pub/plants"
    plant_releases = get_links(url, http)

    # get fungi releases:
    url = "http://ftp.ensemblgenomes.org/pub/fungi"
    fungi_releases = get_links(url, http)

    # get metazoa releases:
    url = "http://ftp.ensemblgenomes.org/pub/metazoa"
    metazoa_releases = get_links(url, http)

    resources = importlib_resources.files("scutls")

    with open(resources / "assets" / "genome_ensembl_release_all.txt", "w") as f:
        _tem_vet = [x.split("-")[1] for x in vet_releases]
        _tem_plant = [x.split("-")[1] for x in plant_releases]
        _tem_fungi = [x.split("-")[1] for x in fungi_releases]
        _tem_metazoa = [x.split("-")[1] for x in metazoa_releases]
        f.write("vertebrates: " + ", ".join(_tem_vet) + "\n")
        f.write("plants: " + ", ".join(_tem_plant) + "\n")
        f.write("fungi: " + ", ".join(_tem_fungi) + "\n")
        f.write("metazoa: " + ", ".join(_tem_metazoa) + "\n")

def get_ensembl_url_json(vertebrates, plants, fungi, metazoa):
    """
    Extract URLs given 4 release numbers via BS.
    """
    genome_download_links = {}
    http = urllib3.PoolManager()

    ## vertebrates
    ftp_site = "http://ftp.ensembl.org/pub/"
    vertebrate_release = "release-104"
    genome_download_links = get_download_urls(http, ftp_site, vertebrates, genome_download_links)

    ## plants
    ftp_site = "http://ftp.ensemblgenomes.org/pub/plants/"
    plant_release = "release-51"
    genome_download_links = get_download_urls(http, ftp_site, plants, genome_download_links)

    ## fungi
    ftp_site = "http://ftp.ensemblgenomes.org/pub/fungi/"
    fungi_release = "release-51"
    genome_download_links = get_download_urls(http, ftp_site, fungi, genome_download_links)

    ## other metazoa
    ftp_site = "http://ftp.ensemblgenomes.org/pub/metazoa/"
    metazoa_release = "release-51"
    genome_download_links = get_download_urls(http, ftp_site, metazoa, genome_download_links)

    ## Dump the dict as json:
    resources = importlib_resources.files("scutls")
    _tem = "_".join(["genome_ensembl", vertebrates, plants, fungi, metazoa]) + ".json"
    with open(resources / "assets" / _tem , "w") as outfile:
        json.dump(genome_download_links, outfile)

# _open function that can auto handle both .gz and .fastq
def _open(filename):
    encoding = guess_type(filename)[1]  # uses file extension
    _open = partial(gzip.open, mode='rt') if encoding == 'gzip' else open
    return(_open(filename))
# _open_out function that can handle SeqIO writing to gz or fastq
def _open_out(filename):
    encoding = guess_type(filename)[1]
    _open_out = partial(gzip.open, mode='wt') if encoding == 'gzip' else open
    return(_open_out(filename))

# find the closing number for chunking fastq file
def closest_number(n, m):
    """
    Find the closest number to n that is no less than n and also divisible by m.
    Ref: https://www.geeksforgeeks.org/find-number-closest-n-divisible-m/
    """
    q = int(n / m)
    n1 = m * q
    if((n * m) > 0) :
        n2 = (m * (q + 1))
    else :
        n2 = (m * (q - 1))

    if (abs(n - n1) < abs(n - n2)) :
        if n1 < n:
            n1 += m
        return n1
    if n2 < n:
        n2 += m
    return n2

def chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    # handle edge case: zero lst
    if len(lst) == 0:
        return(lst)
    
    for i in range(0, len(lst), n):
        yield lst[i:i + n]

# obtain the chunk intervals for given fastq and nproc
def fastq_chunk_interval(fastq, nproc = 1):
    with _open(fastq) as f:
        n1 = sum(1 for record in SeqIO.parse(f, "fastq"))
    chunk_size = math.ceil(n1 / nproc)
    chunk_size = closest_number(chunk_size, 4)
    intervals = list(chunks(range(0, n1), chunk_size))
    intervals = {i: intervals[i] for i in range(0, len(intervals))}
    return(intervals)

# obtain fastq that contains specified barcode
def fastq_contain_barcode(interval, fastq, barcode_pattern):
    fastq_hit, fastq_non_hit = [], []
    with _open(fastq) as f:
        for i, record in enumerate(SeqIO.parse(f, "fastq")):
            if i in interval:
                if regex.search(barcode_pattern, str(record.seq)):
                    fastq_hit.append(record)
                else:
                    fastq_non_hit.append(record)
    return([fastq_hit, fastq_non_hit])

# obtain search pattern given string pattern and allowed mismatches (error)
def get_search_pattern(pattern, error, mismatch_only, rc_barcode):
    if mismatch_only:
        error_pattern = "){s<="
    else:
        error_pattern = "){e<="
    
    if not "," in pattern:
        if rc_barcode:
            pattern = str(Seq(pattern).reverse_complement())
        barcode_pattern = "(" + pattern + error_pattern + str(error) + "}"
    else:
        barcode_pattern = ""
        barcodes = pattern.split(",")
        if rc_barcode:
            barcodes = barcodes[::-1]
            
        for barcode in barcodes:
            if rc_barcode:
                pattern = str(Seq(barcode).reverse_complement())
            else:
                pattern = barcode
            barcode_pattern += "(" + pattern + error_pattern + str(error) + "}(.*?)"
    return(barcode_pattern)

# obtain fastq coordinates for specified barcode
def fastq_locate_barcode(interval, fastq, barcode_pattern, pos = 0):
    """
    pos: which match to return: use 0 for the first match, use -1 for the last match.
    The return value will be: record.id match_location_0_based len(record.seq)s
    
    """
    res = []

    with _open(fastq) as f:
        for i, record in enumerate(SeqIO.parse(f, "fastq")):
            if i in interval:
                matches = regex.finditer(barcode_pattern, str(record.seq))
                # tem = [match for match in matches]
                pos_start = [match.start() for match in matches]
                if not len(pos_start) == 0:
                    res.append([record.id, pos_start[pos], len(record.seq)])
                else:
                    res.append([record.id, "NA", len(record.seq)])
    return(res)

# if containing the following characters, the pattern will be passed directly to regex:
special_search_character = ["{", "}", "(", ")", "*", "="]

# map cigar number to letter
cigar_num2char = {}
cigar_num2char[0] = "M"
cigar_num2char[1] = "I"
cigar_num2char[2] = "D"
cigar_num2char[3] = "N"
cigar_num2char[4] = "S"
cigar_num2char[5] = "H"
cigar_num2char[6] = "P"
cigar_num2char[7] = "="
cigar_num2char[8] = "X"

# calculate CIGAR consumption, used for "scutls bam --locate_pos_in_read"
def cigar_consume(cigar_tuple):
    """_calculate the query and reference consumption given cigar_tuple, return tuple (query_consumed, ref_consumed)_
    
    Ref: https://samtools.github.io/hts-specs/SAMv1.pdf
    
    M 0 alignment match (can be a sequence match or mismatch) yes yes
    I 1 insertion to the reference yes no
    D 2 deletion from the reference no yes
    N 3 skipped region from the reference no yes
    S 4 soft clipping (clipped sequences present in SEQ) yes no
    H 5 hard clipping (clipped sequences NOT present in SEQ) no no
    P 6 padding (silent deletion from padded reference) no no
    = 7 sequence match yes yes
    X 8 sequence mismatch yes yes

    Args:
        cigar_tuple (_tuple_): _CIGAR tuple_
    """
    
    if cigar_tuple[0] == 0:
        return((cigar_tuple[1], cigar_tuple[1]))
    elif cigar_tuple[0] == 1: 
        return((cigar_tuple[1], 0))
    elif cigar_tuple[0] == 2: 
        return((0, cigar_tuple[1]))
    elif cigar_tuple[0] == 3: 
        return((0, cigar_tuple[1]))
    elif cigar_tuple[0] == 4: 
        return((cigar_tuple[1], 0))
    elif cigar_tuple[0] == 5: 
        return((0, 0))
    elif cigar_tuple[0] == 6: 
        return((0, 0))
    elif cigar_tuple[0] == 7: 
        return((cigar_tuple[1], cigar_tuple[1]))
    elif cigar_tuple[0] == 8:
        return((cigar_tuple[1], cigar_tuple[1]))
    else:
        print("wrong CIGAR!")
        exit()

# obtain bam chunk intervals for multiprocessing
def bam_chunk_interval(bam, nproc = 1):
    index_file_path = bam + '.bai'
    if not os.path.exists(index_file_path):
        pysam.index(bam)
    with pysam.AlignmentFile(bam, 'rb') as bam_file:
        n1 = sum(1 for alignment in bam_file)
        
    chunk_size = math.ceil(n1 / nproc)
    intervals = list(chunks(range(0, n1), chunk_size))
    intervals = {i: intervals[i] for i in range(0, len(intervals))}
    return(intervals)

# locate the read site position for given ref coordinate for each interval
def bam_locate_pos_in_read(interval, bam, ref_coordinate):
    res = []
    with pysam.AlignmentFile(bam, 'rb') as bam_file:
        for i, alignment in enumerate(bam_file):
            if i in interval:
                if alignment.reference_start > ref_coordinate:
                    # alignment does not span ref_pos
                    res.append(alignment.query_name + ",NA")
                elif alignment.reference_start == ref_coordinate:
                    res.append(alignment.query_name + "," + str(alignment.query_alignment_start))
                else:
                    ref_pos  = alignment.reference_start
                    # read_pos = alignment.query_alignment_start # soft clip also consume query, will double count if set like this
                    read_pos = 0
                    for i, v in enumerate(alignment.cigartuples):
                        query_consume, ref_consume = cigar_consume(v)
                        ref_pos_tem = ref_pos + ref_consume 
                        read_pos_tem = read_pos + query_consume

                        if ref_pos_tem < ref_coordinate:
                            ref_pos = ref_pos_tem
                            read_pos = read_pos_tem
                        elif ref_pos_tem >= ref_coordinate:
                            if read_pos == read_pos_tem:
                                res.append(alignment.query_name + "," + str(read_pos))
                                break
                            else:
                                if ref_pos_tem - ref_pos == read_pos_tem - read_pos:
                                    # the last checked CIGAR tuple must be M
                                    read_pos_site = read_pos + (ref_coordinate - ref_pos)
                                    assert read_pos_site == read_pos_tem - (ref_pos_tem - ref_coordinate), "error 1"
                                    res.append(alignment.query_name + "," + str(read_pos_site))
                                    break
                                else:
                                    # the last checked CIGAR tuple must be D, and already covered since read_pos must == read_pos_tem
                                    res.append(alignment.query_name + ",invalid")
                                    break
    return(res)

# locate for each cigar tuple the corresponding reference coordinate for each read
def bam_locate_pos_in_ref(interval, bam):
    res = []
    with pysam.AlignmentFile(bam, 'rb') as bam_file:
        for i, alignment in enumerate(bam_file):
            if i in interval:
                ref_pos  = alignment.reference_start
                read_pos = 0
                for i, v in enumerate(alignment.cigartuples):
                    res.append(",".join([alignment.query_name, cigar_num2char[v[0]], str(v[1]), str(ref_pos)]))
                    
                    query_consume, ref_consume = cigar_consume(v)
                    ref_pos = ref_pos + ref_consume 
                    read_pos = read_pos + query_consume
    return(res)


if __name__ == "__main__":
    update_ensembl_release()



import sys
from argparse import ArgumentParser
from scutls import cli
from scutls.__init__ import __version__

parser = ArgumentParser(description = "Single-cell sequencing utility tool")
parser.add_argument("-v", "--version", action="version", version="%(prog)s " + str(__version__))

subparsers = parser.add_subparsers(title = "Subcommands")

# subcommand: download
parser_download = subparsers.add_parser(
    "download", description = "download genome/annotation file from UCSC/ENSEMBL")
parser_download.add_argument(
    "-o", "--outdir",
    help = "output directory, default to ./",
    required = False, type = str, default = "./"
)
parser_download.add_argument(
    "-lgu",
    "--list_genome_ucsc",
    help = "list all available UCSC genome names",
    required = False,
    default = False,
    action='store_true' # quick hack to skip requred input: https://stackoverflow.com/questions/59363298/argparse-expected-one-argument
)
parser_download.add_argument(
    "-gu",
    "--genome_ucsc",
    help = "a UCSC genome name to download",
    required = False,
    default = None,
    type = str
)
parser_download.add_argument(
    "-lge",
    "--list_genome_ensembl",
    help = "list all available ENSEMBL genome names",
    required = False,
    default = False,
    action='store_true' # quick hack to skip requred input: https://stackoverflow.com/questions/59363298/argparse-expected-one-argument
)
parser_download.add_argument(
    "-ge",
    "--genome_ensembl",
    help = "an ENSEMBL genome name to download",
    required = False,
    default = None,
    type = str
)
parser_download.add_argument(
    "-au",
    "--annotation_ucsc",
    help = "an UCSC genome name to download annotation file",
    required = False,
    default = None,
    type = str
)
parser_download.add_argument(
    "-ae",
    "--annotation_ensembl",
    help = "an ENSEMBL genome name to download annotation file",
    required = False,
    default = None,
    type = str
)
parser_download.add_argument(
    "-er",
    "--ensembl_release",
    help = "list all ENSEMBL releases and the one in use",
    required = False,
    default = False,
    action='store_true' # quick hack to skip requred input: https://stackoverflow.com/questions/59363298/argparse-expected-one-argument
)
parser_download.add_argument(
    "-eru",
    "--ensembl_release_use",
    help = "update ENSEMBL release in use, input 4 release numbers separated by comma (vetebrate, plants, fungi, metazoa), e.g. -eru '104, 51, 51, 51'",
    required = False,
    default = None,
    type = str
)
parser_download.add_argument(
    "-erup",
    "--ensembl_release_update",
    help = "update ENSEMBL releases",
    required = False,
    default = False,
    action='store_true' # quick hack to skip requred input: https://stackoverflow.com/questions/59363298/argparse-expected-one-argument
)
parser_download.set_defaults(func = cli.run_download)

# sub-command: fastq
parser_fastq = subparsers.add_parser(
    "fastq", description = "process fastq files")
parser_fastq.add_argument(
    "-i", "--input",
    help = "input fastq, can end with .gz",
    required = True, type = str, default = None
)
parser_fastq.add_argument(
    "-o", "--output",
    help = "output fastq name, can end with .gz",
    required = True, type = str, default = None
)
## add mutually exclusive groups
parser_fastq_group = parser_fastq.add_mutually_exclusive_group()
parser_fastq_group.add_argument(
    "-u", "--unique",
    help = "keep unique records (by sequence ID) in fastq file",
    required = False, default = False, action = "store_true"
)
parser_fastq_group.add_argument(
    "-j", "--join",
    help = "append each read from --join fastq to corresponding line in --input fastq, not 'cat'; the read count and id must be the same in the two fastq files",
    required = False, type = str, default = None
)
parser_fastq_group.add_argument(
    "-fq", "--filter_q",
    help = "filter reads based on mean read quality cutoff (--filter_q). Default to 25. Reads failed to pass this filter will be saved into no_pass_xxx.fastq.gz where xxx represents the name specified with --output flag.",
    required = False, type = int, default = 25
)
## set default values
parser_fastq.set_defaults(func = cli.run_fastq)

# sub-command: bam
parser_bam = subparsers.add_parser(
    "bam", description = "process bam files")
parser_bam.add_argument(
    "-i", "--input",
    help = "input bam, must end with .bam",
    required = True, type = str, default = None
)
parser_bam.add_argument(
    "-o", "--output",
    help = "output file name",
    required = False, type = str, default = None
)
parser_bam.add_argument(
    "-nproc", "--num_processor",
    help = "number of parallel jobs",
    required = False, type = int, default = 1
)
## add mutually exclusive groups
parser_bam_group = parser_bam.add_mutually_exclusive_group()
parser_bam_group.add_argument(
    "-lpir", "--locate_pos_in_read",
    help = "reference coordinate to be located in each read",
    required = False, type = int, default = None
)
parser_bam_group.add_argument(
    "-lpiref", "--locate_pos_in_ref",
    help = "find out the coordinates in reference for each cigar tuple for each read, output will be like: read_id,cigar_char,cigar_char_length,ref_pos",
    required = False, action = 'store_true', default = False
)
## set default values
parser_bam.set_defaults(func = cli.run_bam)

# sub-command: barcode
parser_barcode = subparsers.add_parser(
    "barcode", description = "handle barcode file")
parser_barcode.add_argument(
    "-i", "--input",
    help = "input fastq, can end with .gz",
    required = True, type = str, default = None)
parser_barcode.add_argument(
    "-o", "--output",
    help = "output fastq name, can end with .gz",
    required = False, type = str, default = None
)
parser_barcode.add_argument(
    "-o2", "--output2",
    help = "output fastq name for non-hit reads, can end with .gz",
    required = False, type = str, default = None
)
## add mutually exclusive groups
parser_barcode_group = parser_barcode.add_mutually_exclusive_group()
parser_barcode_group.add_argument(
    "-c", "--contain",
    help = "barcode string to search against the fastq, seperate by comma if multiple. e.g. 'AACCC,GTCCC,CAAA'",
    required = False, type = str, default = None
)
parser_barcode_group.add_argument(
    "-l", "--locate",
    help = "barcode string to search against the fastq, seperate by comma if multiple. e.g. 'AACCC,GTCCC,CAAA'",
    required = False, type = str, default = None
)
parser_barcode.add_argument(
    "-p", "--position",
    help = "return which matched location, default to 0 meaning that the first matched location, use -1 for that last matched location",
    required = False, type = int, default = 0
)

parser_barcode.add_argument(
    "-e", "--error",
    help = "allowed mismatchs (including INDELs) when searching for barcode string",
    required = False, type = int, default = 1
)
parser_barcode.add_argument(
    "-m", "--mismatch_only",
    help = "exclude INDELs when searching for barcode string",
    required = False, default = False, 
    action = 'store_true'
)
parser_barcode.add_argument(
    "-rcb", "--rc_barcode", 
    help = "reverse complement the barcode before searching",
    required = False,
    default = False,
    action = 'store_true' # quick hack to skip requred input: https://stackoverflow.com/questions/59363298/argparse-expected-one-argument
)
parser_barcode.add_argument(
    "-nproc", "--num_processor",
    help = "number of parallel jobs",
    required = False, type = int, default = 1
)
## set default values
parser_barcode.set_defaults(func = cli.run_barcode)

def main():
    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()
    else:
        args = parser.parse_args()
        args.func(args) # parse the args and call whatever function was selected

if __name__ == "__main__":
    main()



from scutls import download, fastq, barcode, bam

# the default values are specified via the arguments.py file
# should use full argument name for both left and right side

def run_download(args):
    download.download(
    list_genome_ucsc = args.list_genome_ucsc,
    genome_ucsc = args.genome_ucsc,
    list_genome_ensembl = args.list_genome_ensembl,
    genome_ensembl = args.genome_ensembl,
    annotation_ucsc = args.annotation_ucsc,
    annotation_ensembl = args.annotation_ensembl,
    ensembl_release = args.ensembl_release,
    ensembl_release_use = args.ensembl_release_use,
    ensembl_release_update = args.ensembl_release_update,
    outdir = args.outdir
    )

def run_fastq(args):
    fastq.fastq(
    input  = args.input,
    output = args.output,
    unique = args.unique,
    join   = args.join,
    filter_q = args.filter_q
    )

def run_bam(args):
    bam.bam(
    input  = args.input,
    output = args.output,
    locate_pos_in_read = args.locate_pos_in_read,
    locate_pos_in_ref = args.locate_pos_in_ref,
    nproc = args.num_processor
    )

def run_barcode(args):
    barcode.barcode(
    input  = args.input,
    output = args.output,
    output2 = args.output2,
    contain = args.contain,
    locate = args.locate,
    pos = args.position,
    error   = args.error,
    mismatch_only = args.mismatch_only,
    rc_barcode = args.rc_barcode,
    nproc = args.num_processor # must use full name
    )


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['scutls']

package_data = \
{'': ['*'], 'scutls': ['assets/*']}

install_requires = \
['GEOparse>=2.0.3,<3.0.0',
 'biopython>=1.80,<2.0',
 'bs4>=0.0.1,<0.0.2',
 'importlib-resources>=5.8.0,<6.0.0',
 'pysam>=0.21.0,<0.22.0',
 'regex>=2023.5.5,<2024.0.0',
 'urllib3>=1.26.10,<2.0.0',
 'wget>=3.2,<4.0']

entry_points = \
{'console_scripts': ['scutls = scutls.arguments:main']}

setup_kwargs = {
    'name': 'scutls',
    'version': '0.3.3',
    'description': 'Single-cell data processing utility tools',
    'long_description': 'None',
    'author': 'Kai Hu',
    'author_email': 'kai.hu@umassmed.edu',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import argparse
import functools
from multiprocessing import Pool
from Bio import SeqIO, bgzf
from Bio.Seq import Seq
from .util import fastq_chunk_interval, fastq_contain_barcode, get_search_pattern, fastq_locate_barcode, special_search_character
import os
import regex

def barcode(input = None, output = None, output2 = None, contain = None, locate = None, pos = 0, error = 1, mismatch_only = False, rc_barcode = False, nproc = 1):
    """barcode subcommand
    Paramters
    ---------

    input : str
        input file name, auto detects .gz
    output : str
        output file name, auto detects .gz, contains fastq that contains specified barcode via contain
    contain: str
        barcode to detect, if multiple barcodess, separate with comma: "AATTCCC,AGGGCCC,CCGGCG", if pattern contains special characters including "()", "{}", "*", "=", the pattern will be passed as searching pattern
        # optional for "contain" sub-command:
        error: int
            allowed nucleotide mismatches (can be INDELs) when searching regex pattern, default to 1
        mismatch_only: bool
            if true, only mismatches (INDELs excluded) are allowed when searching regex pattern, default to False
        nproc: int
            number of parallel jobs, default to 1
        rc_barcode: bool
            take reverse complement of the barcode when aligning, default to False
    locate: str
        barcode to detect its location in reads
        # shared arguments with "contain": error, mismatch_only, nproc, rc_barcode
        pos: int
            which matched position to return, default to 0 meaning the first, use -1 to indicate the last.
        
    """

    args = list(locals().keys())

    local = locals()
    if all(bool(local[key]) is not True for key in args): 
        print("scutls fastq: warning: use 'scutls contain -h' for usage")
    # use True since args can be either None or False
    # arguments.py defines requirments of each argument

    # check if input fastq contains specified barcode:
    if contain:
        # prepare search pattern
        for character in special_search_character:
            if character in contain:
                barcode_pattern = contain 
                break
            else:
                barcode_pattern = get_search_pattern(pattern = contain, error = error, mismatch_only = mismatch_only, rc_barcode = rc_barcode)
        
        # print("barcode_pattern: ", barcode_pattern)

        # multiprocessing
        if not output == None:
            print("Saving to " + output + " ...")
        else:
            print("Processing ...")
        intervals = fastq_chunk_interval(input, nproc = nproc)
        p = Pool(nproc)
        res = p.map_async(
            functools.partial(
                fastq_contain_barcode,
                fastq = input,
                barcode_pattern = barcode_pattern),
                intervals.values()).get()
        fastq_hit, fastq_non_hit = [], []
        
        for x in range(len(res)):
            fastq_hit = fastq_hit + [hit for hit in res[x][0]]
            fastq_non_hit = fastq_non_hit + [non_hit for non_hit in res[x][1]]

        if not output == None:
            if not os.path.dirname(output) == "":
                os.makedirs(os.path.dirname(output), exist_ok=True)
            if output.endswith(".gz"):
                with bgzf.BgzfWriter(output, "wb") as outgz:
                    SeqIO.write(sequences = fastq_hit, handle = outgz, format="fastq")
            else:
                SeqIO.write(fastq_hit, output, "fastq")
        else:
            for record in fastq_hit:
                print(record.seq, [x.start() for x in regex.finditer(barcode_pattern, str(record.seq))])
        
        if not output2 == None:
            print("Saving to " + output2 + " ...")
            if not os.path.dirname(output2) == "":
                os.makedirs(os.path.dirname(output2), exist_ok=True)
            if output2.endswith(".gz"):
                with bgzf.BgzfWriter(output2, "wb") as outgz:
                    SeqIO.write(sequences = fastq_non_hit, handle = outgz, format="fastq")
            else:
                SeqIO.write(fastq_non_hit, output2, "fastq")
        
        print("Done!")

    # check if input fastq contains specified barcode
    if locate:
        # prepare search pattern
        for character in special_search_character:
            if character in locate:
                barcode_pattern = locate 
                break
        else:
            barcode_pattern = get_search_pattern(pattern = locate, error = error, mismatch_only = mismatch_only, rc_barcode = rc_barcode)

        # multiprocessing
        intervals = fastq_chunk_interval(input, nproc = nproc)
        p = Pool(nproc)
        res_list = p.map_async(
            functools.partial(
                fastq_locate_barcode,
                fastq = input,
                barcode_pattern = barcode_pattern,
                pos = pos),
                intervals.values()).get()
        for res in res_list:
            for res_items in res:
                print(res_items[0], res_items[1], res_items[2])


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--input',  '-i', type   = str)
    parser.add_argument('--output', '-o', type   = str)
    parser.add_argument('--output2', '-o2', type = str)
    parser.add_argument('--contain', '-c', type = str)
    parser.add_argument('--locate', '-l', type = str)
    parser.add_argument('--position', '-p', type = int)
    
    parser.add_argument('--error',   '-e', type   = int)
    parser.add_argument('--mismatch_only', '-m', default = False)
    parser.add_argument('--rc_barcode', '-rcb', default = False)
    parser.add_argument('--num_processor', '-nproc', default = 1)
    
if __name__ == "__main__":
    main()



import json
import importlib_resources
import wget
import os
from os import path
import cmd
import argparse
from .util import update_ensembl_release, get_ensembl_url_json

def download(list_genome_ucsc = False, list_genome_ensembl = False, genome_ucsc = None, genome_ensembl = None, annotation_ucsc = None, annotation_ensembl = None, ensembl_release = False, ensembl_release_use = None, ensembl_release_update = False, outdir = "./"):
    """download subcommand
    Paramters
    ---------

    list_genome_ucsc : bool
        list all UCSC genome names
    list_genome_ensembl : bool
        list all ENSEMBL genome names
    genome_ucsc : str
        download genome given UCSC genome name
    genome_ensembl : str
        download genome given ENSEMBL genome name
    outdir : str
        output directory, default to "./"
    annotation_ucsc : str
        download annotation file given UCSC genome name
    annotation_ensembl : str
        download annotation file given ENSEMBL genome name
    ensembl_release : bool
        list all ENSEMBL releases and the one in use
    ensembl_release_use : str
        update ENSEMBL release in use, input 4 release numbers separated by comma (vertebrates, plants, fungi, metazoa), e.g. -eru '104, 51, 51, 51'
    ensembl_release_update : bool
        update genome_ensembl_release_all.txt

    """

    args = list(locals().keys())
    args.remove("outdir")

    local = locals()
    if all(bool(local[key]) is not True for key in args): # use True since args can be either None or False
        print("scutls download: warning: use 'scutls download -h' for usage")

    resources = importlib_resources.files("scutls")

    # update ensembl_release_all.txt file
    if ensembl_release_update:
        try:
            print("Updating releases in cache ...")
            update_ensembl_release()
            print("Done! Use --ensembl_release to see full list.")
        except:
            print("Updating failed! Pls try again later!")

    # update ensembl_release_use:
    if not ensembl_release_use == None:
        try:
            with open(resources / "assets" / "genome_ensembl_release_use.txt", "w") as f:
                _tem = [x.strip() for x in ensembl_release_use.split(",")]
                assert len(_tem) == 4, "Must provide 4 release numbers!"
                f.write("vertebrates: " + _tem[0] + "\n")
                f.write("plants: " + _tem[1] + "\n")
                f.write("fungi: " + _tem[2] + "\n")
                f.write("metazoa: " + _tem[3] + "\n")
                print("ENSEMBL release in use updated to: " + ",".join(_tem) + "!")
        except:
            print("Updating failed! Check out the correct input format!")

    dict_genome_ucsc = json.loads((resources / "assets" / "genome_ucsc.json").read_bytes()) # https://stackoverflow.com/questions/6028000/how-to-read-a-static-file-from-inside-a-python-package

    with open(resources / "assets" / "genome_ensembl_release_use.txt") as f:
        ensembl_release_current = {}
        for line in f:
            key, value = line.split(": ")
            ensembl_release_current[key] = value.strip()
    with open(resources / "assets" / "genome_ensembl_release_all.txt") as f:
        ensembl_release_all = {}
        for line in f:
            key, values = line.split(": ")
            ensembl_release_all[key] = values
    _tem = "genome_ensembl_" + "_".join(ensembl_release_current.values()) + ".json"

    # print all ensembl releases and the one in use:
    if ensembl_release:
        _tem = ", ".join(ensembl_release_all)
        print("All ENSEMBL releases: ")
        for key in ensembl_release_all:
            print("\t", key, ":", ensembl_release_all[key], end = "")
        print("\nENSEMBL relsease in use: ")
        for key in ensembl_release_current:
            print("\t", key, ":", ensembl_release_current[key], end = "\n")

    try:
        _tem_current = "_".join(["genome_ensembl", *list(ensembl_release_current.values())]) + ".json"
        dict_genome_ensembl = json.loads((resources / "assets" / _tem_current).read_bytes())
    except:
        print("ENSEMBL release " + ",".join(ensembl_release_current.values()) + " not in cache!")
        print("Creating caching file ... may take several minutes ...")
        try:
            get_ensembl_url_json(*list(ensembl_release_current.values()))
            print("Success!")
        except:
            print("Failed! Please try again later")

    # print available UCSC genomes:
    if list_genome_ucsc:
        print("Supported UCSC genomes:")
        cli = cmd.Cmd()
        cli.columnize(list(dict_genome_ucsc.keys()), displaywidth=80)

    # print available ENSEMBL genomes:
    if list_genome_ensembl:
        print("Supported ENSEMBL genomes:")
        cli = cmd.Cmd()
        cli.columnize(list(dict_genome_ensembl.keys()), displaywidth=80)

    # download specified UCSC genome:
    if type(genome_ucsc) != type(None):
        if not genome_ucsc in dict_genome_ucsc.keys():
            print("WARNING: " + genome_ucsc + " not supported!")
        else:
            url = dict_genome_ucsc[genome_ucsc]["genome"]
            # filename = wget.download(url)
            print("Downloading " + genome_ucsc + " ...")
            if not os.path.exists(outdir):
                os.mkdir(outdir)
            wget.download(url, out = outdir)
            print("\nDownloaded to " + path.join(outdir, path.basename(url)) + "!")

    # download specified ENSEMBL genome:
    if type(genome_ensembl) != type(None):
        if not genome_ensembl in dict_genome_ensembl.keys():
            print("WARNING: " + genome_ensembl + " not supported!")
        else:
            url = dict_genome_ensembl[genome_ensembl]["genome"]
            print("Downloading " + genome_ensembl + " ...")
            if not os.path.exists(outdir):
                os.mkdir(outdir)
            wget.download(url, out = outdir)
            print("\nDownloaded to " + path.join(outdir, path.basename(url)) + "!")

    # download specified UCSC annotation file:
    if type(annotation_ucsc) != type(None):
        if not annotation_ucsc in dict_genome_ucsc.keys():
            print("WARNING: " + annotation_ucsc + " not supported!")
        else:
            url = dict_genome_ucsc[annotation_ucsc]["gtf"]
            print("Downloading annotation file: " + annotation_ucsc + " ...")
            if not os.path.exists(outdir):
                os.mkdir(outdir)
            wget.download(url, out = outdir)
            print("\nDownloaded to " + path.join(outdir, path.basename(url)) + "!")

    # download specified ENSEMBL annotation file:
    if type(annotation_ensembl) != type(None):
        if not annotation_ensembl in dict_genome_ensembl.keys():
            print("WARNING: " + annotation_ensembl + " not supported!")
        else:
            url = dict_genome_ensembl[annotation_ensembl]["gtf"]
            print("Downloading annotatiion file: " + annotation_ensembl + " ...")
            if not os.path.exists(outdir):
                os.mkdir(outdir)
            wget.download(url, out = outdir)
            print("\nDownloaded to " + path.join(outdir, path.basename(url)) + "!")


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--list_genome_ucsc', '-lgu', action='store_true')
    parser.add_argument('--list_genome_ensembl', '-lge', action='store_true')
    parser.add_argument('--genome_ucsc', '-gu', type = str)
    parser.add_argument('--genome_ensembl', '-ge', type = str)

    args = parser.parse_args()
    download(args.list_genome_ucsc, args.list_genome_ensembl, args.genome_ucsc, args.genome_ensembl, outdir = "./")

if __name__ == "__main__":
    main()
